import datetime, pandas, numpy, os, pickle, sys
import sklearn, sklearn.decomposition
import scipy, scipy.stats
import multiprocessing
from pkg_resources import Requirement, resource_filename
import logging
import json
from collections import defaultdict

from miner2 import util, coexpression


def axis_tfs(axesDf, tfList, expressionData, correlation_threshold=0.3):
    axesArray = numpy.array(axesDf.T)
    if correlation_threshold > 0:
        tfArray = numpy.array(expressionData.reindex(tfList))
    axes = numpy.array(axesDf.columns)
    tfDict = {}

    if type(tfList) is list:
        tfs = numpy.array(tfList)
    elif type(tfList) is not list:
        tfs = np.array(list(tfList))

    if correlation_threshold == 0:
        for axis in range(axesArray.shape[0]):
            tfDict[axes[axis]] = tfs
        return tfDict

    for axis in range(axesArray.shape[0]):
        tfDict_key = axes[axis]
        # WW: the problem here is that tfCorrelation can have NaN values which warnings
        # are shown for
        tfCorrelation = coexpression.pearson_array(tfArray,axesArray[axis,:])
        #print('tfCorrelation: ' + str(numpy.abs(tfCorrelation)))
        #print('correlation.threshold: ' + str(correlation_threshold))
        tfDict[tfDict_key] = tfs[numpy.where(numpy.abs(tfCorrelation) >= correlation_threshold)[0]]
    return tfDict



def enrichment(axes, revised_clusters, expression_data, correlation_threshold=0.3,
               num_cores=1, p=0.05,
               database="tfbsdb_tf_to_genes.pkl",
               database_path=None):

    logging.info("mechanistic inference")

    if database_path is None:
        tf_2_genes_path = resource_filename(Requirement.parse("miner2"),
                                            'miner2/data/{}'.format(database))
    else:
        tf_2_genes_path = database_path

    with open(tf_2_genes_path, 'rb') as f:
        tfToGenes = pickle.load(f)

    if correlation_threshold <= 0:
        allGenes = [int(len(expression_data.index))]
    elif correlation_threshold > 0:
        allGenes = list(expression_data.index)

    tfs = list(tfToGenes.keys())
    tfMap = axis_tfs(axes, tfs, expression_data, correlation_threshold=correlation_threshold)
    taskSplit = util.split_for_multiprocessing(list(revised_clusters.keys()), num_cores)
    tasks = [[taskSplit[i], (allGenes, revised_clusters, tfMap, tfToGenes, p)]
             for i in range(len(taskSplit))]
    tfbsdbOutput = multiprocess(tfbsdb_enrichment, tasks)
    return condense_output(tfbsdbOutput)


def multiprocess(function,tasks):
    import multiprocessing, multiprocessing.pool
    hydra=multiprocessing.pool.Pool(len(tasks))  
    output=hydra.map(function,tasks)   
    hydra.close()
    hydra.join()
    return output


def hyper(population,set1,set2,overlap):

    b = max(set1,set2)
    c = min(set1,set2)
    hyp = scipy.stats.hypergeom(population,b,c)
    prb = sum([hyp.pmf(l) for l in range(overlap,c+1)])

    return prb


def get_principal_df(dict_, expressionData, regulons=None, subkey='genes',
                     min_number_genes=8, random_state=12):
    pcDfs = []
    setIndex = set(expressionData.index)

    if regulons is not None:
        dict_, df = regulonDictionary(regulons)
    for i in list(dict_.keys()):
        if subkey is not None:
            genes = list(set(dict_[i][subkey])&setIndex)
            if len(genes) < min_number_genes:
                continue
        elif subkey is None:
            genes = list(set(dict_[i])&setIndex)
            if len(genes) < min_number_genes:
                continue

        pca = sklearn.decomposition.PCA(1, random_state=random_state)
        principalComponents = pca.fit_transform(expressionData.loc[genes,:].T)
        principalDf = pandas.DataFrame(principalComponents)
        principalDf.index = expressionData.columns
        principalDf.columns = [str(i)]

        normPC = numpy.linalg.norm(numpy.array(principalDf.iloc[:,0]))
        pearson = scipy.stats.pearsonr(principalDf.iloc[:,0],
                                       numpy.median(expressionData.loc[genes,:], axis=0))
        signCorrection = pearson[0] / numpy.abs(pearson[0])
        principalDf = signCorrection * principalDf / normPC

        pcDfs.append(principalDf)

    principalMatrix = pandas.concat(pcDfs, axis=1)
    return principalMatrix


def get_regulon_dictionary(regulons):
    regulonModules = {}
    df_list = []

    for tf in list(regulons.keys()):
        for key in list(regulons[tf].keys()):
            genes = regulons[tf][key]
            id_ = str(len(regulonModules))
            regulonModules[id_] = regulons[tf][key]
            for gene in genes:
                df_list.append([id_,tf,gene])

    array = numpy.vstack(df_list)
    df = pandas.DataFrame(array)
    df.columns = ["Regulon_ID", "Regulator", "Gene"]

    return regulonModules, df


def condense_output(output,output_type=dict):
    if output_type is dict:
        results = {}
        for i in range(len(output)):
            resultsDict = output[i]
            keys = list(resultsDict.keys())
            for j in range(len(resultsDict)):
                key = keys[j]
                results[key] = resultsDict[key]
        return results
    elif output_type is not dict:
        results = pandas.concat(output, axis=0)

    return results


def tfbsdb_enrichment(task):
    start, stop = task[0]
    allGenes, revisedClusters, tfMap, tfToGenes, p = task[1]
    keys = list(revisedClusters.keys())[start:stop]

    if len(allGenes) == 1:

        population_size = int(allGenes[0])
        clusterTfs = {}
        for key in keys:
            for tf in tfMap[str(key)]:
                hits0TfTargets = tfToGenes[tf]
                hits0clusterGenes = revisedClusters[key]
                overlapCluster = list(set(hits0TfTargets)&set(hits0clusterGenes))
                if len(overlapCluster) <= 1:
                    continue
                pHyper = hyper(population_size,len(hits0TfTargets),len(hits0clusterGenes),len(overlapCluster))
                if pHyper < p:
                    if key not in list(clusterTfs.keys()):
                        clusterTfs[key] = {}
                    clusterTfs[key][tf] = [pHyper,overlapCluster]

    elif len(allGenes) > 1:
        population_size = len(allGenes)
        clusterTfs = {}
        for key in keys:
            for tf in tfMap[str(key)]:
                hits0TfTargets = list(set(tfToGenes[tf])&set(allGenes))
                hits0clusterGenes = revisedClusters[key]
                overlapCluster = list(set(hits0TfTargets)&set(hits0clusterGenes))
                if len(overlapCluster) <= 1:
                    continue
                pHyper = hyper(population_size,len(hits0TfTargets),len(hits0clusterGenes),len(overlapCluster))
                if pHyper < p:
                    if key not in list(clusterTfs.keys()):
                        clusterTfs[key] = {}
                    clusterTfs[key][tf] = [pHyper,overlapCluster]

    return clusterTfs


def get_coregulation_modules(mechanistic_output):
    coregulation_modules = {}
    for i in mechanistic_output.keys():
        for key in mechanistic_output[i].keys():
            if key not in coregulation_modules.keys():
                coregulation_modules[key] = {}
            genes = mechanistic_output[i][key][1]
            coregulation_modules[key][i] = genes
    return coregulation_modules


def get_regulons(coregulationModules, min_number_genes=5, freq_threshold=0.333):
    regulons = {}
    keys = list(coregulationModules.keys())
    for i in range(len(keys)):
        tf = keys[i]
        normDf = coincidence_matrix(coregulationModules, key=i, freq_threshold=freq_threshold)
        unmixed = coexpression.unmix(normDf)
        remixed = coexpression.remix(normDf, unmixed)
        if len(remixed) > 0:
            for cluster in remixed:
                if len(cluster) >= min_number_genes:
                    if tf not in list(regulons.keys()):
                        regulons[tf] = {}
                    regulons[tf][len(regulons[tf])] = cluster
    return regulons


def coincidence_matrix(coregulationModules, key, freq_threshold=0.333):
    tf = list(coregulationModules.keys())[key]
    subRegulons = coregulationModules[tf]
    srGenes = list(set(numpy.hstack([subRegulons[i] for i in subRegulons.keys()])))

    template = pandas.DataFrame(numpy.zeros((len(srGenes), len(srGenes))))
    template.index = srGenes
    template.columns = srGenes

    for key in list(subRegulons.keys()):
        genes = subRegulons[key]
        template.loc[genes,genes]+=1
    trace = numpy.array([template.iloc[i,i] for i in range(template.shape[0])]).astype(float)
    normDf = ((template.T)/trace).T
    normDf[normDf < freq_threshold] = 0
    normDf[normDf > 0] = 1

    return normDf


def get_coexpression_modules(mechanistic_output):
    coexpressionModules = {}
    for i in mechanistic_output.keys():
        genes = list(set(numpy.hstack([mechanistic_output[i][key][1]
                                       for key in mechanistic_output[i].keys()])))
        coexpressionModules[i] = genes
    return coexpressionModules


"""
Postprocessing functions
"""
def convert_dictionary(dic, conversion_table):
    conv_dict = defaultdict(list)
    for pref_name, name in conversion_table.iteritems():
        conv_dict[pref_name].append(name)

    converted = defaultdict(list)
    for i in dic.keys():
        genes = dic[i]
        for gene in genes:
            converted[i].extend(conv_dict[gene])
    return converted


def convert_regulons(df, conversion_table):
    """df is dataframe with the columns "Regulon_ID", "Regulator", "Gene"
    """
    reg_ids = []
    regs = []
    genes = []
    conv_dict = defaultdict(list)
    for pref_name, name in conversion_table.iteritems():
        conv_dict[pref_name].append(name)

    for i, row in df.iterrows():
        reg_ids.append(row["Regulon_ID"])
        regs.append(conv_dict[row["Regulator"]][0])
        genes.append(conv_dict[row["Gene"]][0])

    regulon_df_converted = pandas.DataFrame(numpy.vstack([reg_ids, regs, genes]).T)
    regulon_df_converted.columns = ["Regulon_ID","Regulator","Gene"]
    return regulon_df_converted



import datetime,numpy,pandas,time,sys,itertools
import sklearn,sklearn.decomposition
import multiprocessing, multiprocessing.pool
from collections import Counter
import traceback as tb

# Some default constants if the user does not specify any
# default number of iterations for algorithms with iterations
NUM_ITERATIONS = 25
MIN_NUM_GENES = 6
MIN_NUM_OVEREXP_SAMPLES = 4
MAX_SAMPLES_EXCLUDED = 0.5
RANDOM_STATES = 12
OVEREXP_THRESHOLD = 80
NUM_CORES = 1
RECONSTRUCTION_THRESHOLD = 0.925
SIZE_LONG_SET = 50


def cluster(expressionData, min_number_genes=6, min_number_overexp_samples=4,
            max_samples_excluded=0.50, random_state=12, overexpression_threshold=80,
            pct_threshold=80):


    df = expressionData.copy()
    maxStep = int(numpy.round(10 * max_samples_excluded))
    allGenesMapped = []
    bestHits = []

    zero = numpy.percentile(expressionData,0)
    expressionThreshold = numpy.mean([numpy.percentile(expressionData.iloc[:,i][expressionData.iloc[:,i]>zero], overexpression_threshold) for i in range(expressionData.shape[1])])

    startTimer = time.time()
    trial = -1
    for step in range(maxStep):
        trial+=1
        progress = (100./maxStep)*trial
        print('{:.2f} percent complete'.format(progress))
        genesMapped = []
        bestMapped = []

        pca = sklearn.decomposition.PCA(10,random_state=random_state)
        principalComponents = pca.fit_transform(df.T)
        principalDf = pandas.DataFrame(principalComponents)
        principalDf.index = df.columns

        for i in range(10):
            pearson = pearson_array(numpy.array(df), numpy.array(principalDf[i]))
            if len(pearson) == 0:
                continue
            highpass = max(numpy.percentile(pearson,95), 0.1)
            lowpass = min(numpy.percentile(pearson,5), -0.1)
            cluster1 = numpy.array(df.index[numpy.where(pearson > highpass)[0]])
            cluster2 = numpy.array(df.index[numpy.where(pearson < lowpass)[0]])

            for clst in [cluster1,cluster2]:
                pdc = recursive_alignment(clst, expression_data=df,
                                          min_number_genes=min_number_genes,
                                          pct_threshold=pct_threshold)
                if len(pdc)==0:
                    continue
                elif len(pdc) == 1:
                    genesMapped.append(pdc[0])
                elif len(pdc) > 1:
                    for j in range(len(pdc)-1):
                        if len(pdc[j]) > min_number_genes:
                            genesMapped.append(pdc[j])

        allGenesMapped.extend(genesMapped)
        try:
            stackGenes = numpy.hstack(genesMapped)
        except:
            tb.print_exc()  # don't silence exceptions !!
            stackGenes = []
        residualGenes = list(set(df.index)-set(stackGenes))
        df = df.loc[residualGenes,:]

        # computationally fast surrogate for passing the overexpressed significance test:
        for ix in range(len(genesMapped)):
            tmpCluster = expressionData.loc[genesMapped[ix],:]
            tmpCluster[tmpCluster<expressionThreshold] = 0
            tmpCluster[tmpCluster>0] = 1
            sumCluster = numpy.array(numpy.sum(tmpCluster, axis=0))
            numHits = numpy.where(sumCluster > 0.333 * len(genesMapped[ix]))[0]
            bestMapped.append(numHits)
            if len(numHits) > min_number_overexp_samples:
                bestHits.append(genesMapped[ix])

        if len(bestMapped) > 0:
            countHits = Counter(numpy.hstack(bestMapped))
            ranked = countHits.most_common()
            dominant = [i[0] for i in ranked[0:int(numpy.ceil(0.1*len(ranked)))]]
            remainder = [i for i in numpy.arange(df.shape[1]) if i not in dominant]
            df = df.iloc[:,remainder]

    bestHits.sort(key=lambda s: -len(s))

    stopTimer = time.time()
    print('\ncoexpression clustering completed in {:.2f} minutes'.format((stopTimer-startTimer)/60.))
    return bestHits


def combine_clusters(axes, clusters, threshold):
    combine_axes = {}
    filter_keys = numpy.array(list(axes.keys())) # ALO: changed to list because of Py3
    axes_matrix = numpy.vstack([axes[i] for i in filter_keys])
    for key in filter_keys:
        axis = axes[key]
        pearson = pearson_array(axes_matrix,axis)
        combine = numpy.where(pearson > threshold)[0]
        combine_axes[key] = filter_keys[combine]

    revised_clusters = {}
    combined_keys = decompose_dictionary_to_lists(combine_axes)
    for key_list in combined_keys:
        genes = list(set(numpy.hstack([clusters[i] for i in key_list])))
        revised_clusters[len(revised_clusters)] = sorted(genes)

    return revised_clusters


def decompose(geneset, expressionData, minNumberGenes=6, pct_threshold=80):
    fm = make_frequency_matrix(expressionData.loc[geneset,:])
    tst = numpy.multiply(fm,fm.T)
    tst[tst < numpy.percentile(tst,pct_threshold)]=0
    tst[tst > 0]=1
    unmix_tst = unmix(tst)
    unmixedFiltered = [i for i in unmix_tst if len(i)>=minNumberGenes]
    return unmixedFiltered


def decompose_dictionary_to_lists(dict_):
    decomposedSets = []
    for key in list(dict_.keys()):
        newSet = iterative_combination(dict_, key, iterations=NUM_ITERATIONS)
        if newSet not in decomposedSets:
            decomposedSets.append(newSet)
    return decomposedSets


def make_frequency_matrix(matrix, overExpThreshold = 1):

    final_index = None
    if type(matrix) == pandas.core.frame.DataFrame:
        final_index = matrix.index
        matrix = numpy.array(matrix)

    index = numpy.arange(matrix.shape[0])
    matrix[matrix<overExpThreshold] = 0
    matrix[matrix>0] = 1
    frequency_dictionary = {name:[] for name in index}

    for column in range(matrix.shape[1]):
        hits = numpy.where(matrix[:,column]>0)[0]
        geneset = index[hits]
        for name in geneset:
            frequency_dictionary[name].extend(geneset)

    fm = numpy.zeros((len(index),len(index)))
    for key in list(frequency_dictionary.keys()):
        tmp = frequency_dictionary[key]
        if len(tmp) == 0:
            continue
        count = Counter(tmp)
        results_ = numpy.vstack(list(count.items()))
        fm[key,results_[:,0]] = results_[:,1]/float(count[key])

    fm_df = pandas.DataFrame(fm)

    if final_index is not None:
        fm_df.index = final_index
        fm_df.columns = final_index

    return fm_df



def get_axes(clusters, expression_data):
    axes = {}
    for key in clusters.keys():
        genes = clusters[key]
        fpc = sklearn.decomposition.PCA(1)
        principal_components = fpc.fit_transform(expression_data.loc[genes,:].T)
        axes[key] = principal_components.ravel()
    return axes


def gene_mapper(task):
    genes_mapped = []
    df, principal_df, i, min_number_genes = task
    pearson = pearson_array(numpy.array(df), numpy.array(principal_df[i]))
    highpass = max(numpy.percentile(pearson,95), 0.1)
    lowpass = min(numpy.percentile(pearson,5), -0.1)
    cluster1 = numpy.array(df.index[numpy.where(pearson > highpass)[0]])
    cluster2 = numpy.array(df.index[numpy.where(pearson < lowpass)[0]])

    for clst in [cluster1, cluster2]:
        pdc = recursive_alignment(clst, df, min_number_genes)
        if len(pdc) == 0:
            continue
        elif len(pdc) == 1:
            genes_mapped.append(pdc[0])
        elif len(pdc) > 1:
            for j in range(len(pdc)-1):
                if len(pdc[j]) > min_number_genes:
                    genes_mapped.append(pdc[j])

    return genes_mapped


def iterative_combination(dict_, key, iterations=25):
    initial = dict_[key]
    initialLength = len(initial)
    for iteration in range(iterations):
        revised = [i for i in initial]
        for element in initial:
            revised = list(set(revised)|set(dict_[element]))
        revisedLength = len(revised)
        if revisedLength == initialLength:
            return revised
        elif revisedLength > initialLength:
            initial = [i for i in revised]
            initialLength = len(initial)
    return revised



def make_hits_matrix_new(matrix): ### new function developed by Wei-Ju
    num_rows = matrix.shape[0]
    hits_values = numpy.zeros((num_rows,num_rows))

    for column in range(matrix.shape[1]):
        geneset = matrix[:,column]
        hits = numpy.where(geneset > 0)[0]
        rows = []
        cols = []
        cp = itertools.product(hits, hits)
        for row, col in cp:
            rows.append(row)
            cols.append(col)
        hits_values[rows, cols] += 1

    return hits_values


def parallel_overexpress_surrogate(task):
    element, expression_data, expression_threshold = task

    tmp_cluster = expression_data.loc[element, :]
    tmp_cluster[tmp_cluster < expression_threshold] = 0
    tmp_cluster[tmp_cluster > 0] = 1
    sum_cluster = numpy.array(numpy.sum(tmp_cluster, axis=0))
    hits = numpy.where(sum_cluster > 0.333 * len(element))[0]

    return (element, hits)


def pearson_array(array, vector):
    ybar = numpy.mean(vector)
    sy = numpy.std(vector, ddof=1)
    yterms = (vector - ybar) / float(sy)

    array_sx = numpy.std(array, axis=1, ddof=1)

    if 0 in array_sx:
        pass_index = numpy.where(array_sx > 0)[0]
        array = array[pass_index, :]
        array_sx = array_sx[pass_index]

    array_xbar = numpy.mean(array, axis=1)
    product_array = numpy.zeros(array.shape)

    for i in range(0,product_array.shape[1]):
        product_array[:, i] = yterms[i] * (array[:, i] - array_xbar) / array_sx

    return numpy.sum(product_array, axis=1) / float(product_array.shape[1]-1)


def process_coexpression_lists(lists,expressionData,threshold=0.925):
    reconstructed = reconstruction(lists,expressionData,threshold)
    reconstructedList = [reconstructed[i] for i in reconstructed.keys()]
    reconstructedList.sort(key = lambda s: -len(s))
    return reconstructedList


def reconstruction(decomposed_list, expression_data, threshold=RECONSTRUCTION_THRESHOLD):
    clusters = {i:decomposed_list[i] for i in range(len(decomposed_list))}
    axes = get_axes(clusters, expression_data)
    return combine_clusters(axes, clusters, threshold)


def recursive_alignment(geneset, expression_data, min_number_genes=6, pct_threshold=80):
    recDecomp = recursive_decomposition(geneset, expression_data, min_number_genes, pct_threshold)
    if len(recDecomp) == 0:
        return []

    reconstructed = reconstruction(recDecomp,expression_data)
    reconstructedList = [reconstructed[i] for i in list(reconstructed.keys())
                         if len(reconstructed[i]) > min_number_genes]
    reconstructedList.sort(key = lambda s: -len(s))
    return reconstructedList


def recursive_decomposition(geneset, expressionData, minNumberGenes=6, pct_threshold=80):
    unmixedFiltered = decompose(geneset,expressionData,minNumberGenes,pct_threshold)
    if len(unmixedFiltered) == 0:
        return []
    shortSets = [i for i in unmixedFiltered if len(i)<50]
    longSets = [i for i in unmixedFiltered if len(i)>=50]
    if len(longSets)==0:
        return unmixedFiltered
    for ls in longSets:
        unmixedFiltered = decompose(ls,expressionData,minNumberGenes,pct_threshold)
        if len(unmixedFiltered)==0:
            continue
        shortSets.extend(unmixedFiltered)
    return shortSets


def revise_initial_clusters(clusterList, expressionData, threshold=RECONSTRUCTION_THRESHOLD):
    coexpressionLists = process_coexpression_lists(clusterList,expressionData,threshold)
    coexpressionLists.sort(key= lambda s: -len(s))

    for iteration in range(5):
        previousLength = len(coexpressionLists)
        coexpressionLists = process_coexpression_lists(coexpressionLists,expressionData,threshold)
        newLength = len(coexpressionLists)
        if newLength == previousLength:
            break

    coexpressionLists.sort(key= lambda s: -len(s))
    coexpressionDict = {str(i):list(coexpressionLists[i]) for i in range(len(coexpressionLists))}

    return coexpressionDict


def unmix(df, iterations=NUM_ITERATIONS, return_all=False):    
    frequencyClusters = []

    for iteration in range(iterations):
        sumDf1 = df.sum(axis=1)
        maxSum = df.index[numpy.argmax(numpy.array(sumDf1))]
        hits = numpy.where(df.loc[maxSum] > 0)[0]
        hitIndex = list(df.index[hits])
        block = df.loc[hitIndex,hitIndex]
        blockSum = block.sum(axis=1)
        coreBlock = list(blockSum.index[numpy.where(blockSum>=numpy.median(blockSum))[0]])
        remainder = list(set(df.index)-set(coreBlock))
        frequencyClusters.append(coreBlock)
        if len(remainder)==0:
            return frequencyClusters
        if len(coreBlock)==1:
            return frequencyClusters
        df = df.loc[remainder,remainder]
    if return_all:
        frequencyClusters.append(remainder)
    return frequencyClusters


def remix(df,frequencyClusters):
    finalClusters = []
    for cluster in frequencyClusters:
        sliceDf = df.loc[cluster,:]
        sumSlice = sliceDf.sum(axis=0)
        cut = min(0.8,numpy.percentile(sumSlice.loc[cluster]/float(len(cluster)),90))
        minGenes = max(4,cut*len(cluster))
        keepers = list(sliceDf.columns[numpy.where(sumSlice>=minGenes)[0]])
        keepers = list(set(keepers)|set(cluster))
        finalClusters.append(keepers)
        finalClusters.sort(key = lambda s: -len(s))
    return finalClusters



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

PACKAGE_DATA = {
    'miner2': ['data/*']
    }

INSTALL_REQUIRES = ['numpy', 'scipy', 'pandas', 'sklearn', 'lifelines',
                    'matplotlib', 'seaborn', # 'xgboost',
                    'pydot', 'graphviz']
setuptools.setup(
    name="isb_miner2",
    version="0.2.7",
    author="Adrian Lopez Garcia de Lomana",
    author_email="alomana@systemsbiology.org",
    description="A newer version of MINER.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/baliga-lab/miner2",
    packages=['miner2'],
    install_requires = INSTALL_REQUIRES,
    include_package_data=True, package_data=PACKAGE_DATA,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    scripts=['bin/miner2-coexpr', 'bin/miner2-mechinf',
             'bin/miner2-bcmembers', 'bin/miner2-subtypes',
             'bin/miner2-survival', 'bin/miner2-causalinference', 'bin/miner2-causalinf-pre',
             'bin/miner2-causalinf-post', 'bin/miner2-neo', 'bin/miner2-riskpredict'])



import pandas
import numpy
import pickle
import os
from scipy import stats

from miner2 import subtypes
from miner2 import util


def _bicluster_tf_incidence(mechanisticOutput, regulons=None):

    if regulons is not None:
        allTfs = list(regulons.keys())

        tfCount = []
        ct = 0
        for tf in regulons.keys():
            tfCount.append([])
            for key in regulons[tf].keys():
                tfCount[-1].append(str(ct))
                ct+=1

        allBcs = numpy.hstack(tfCount)

        bcTfIncidence = pandas.DataFrame(numpy.zeros((len(allBcs), len(allTfs))))
        bcTfIncidence.index = allBcs
        bcTfIncidence.columns = allTfs

        for i in range(len(allTfs)):
            tf = allTfs[i]
            bcs = tfCount[i]
            bcTfIncidence.loc[bcs,tf] = 1

        index = numpy.sort(numpy.array(bcTfIncidence.index).astype(int))
        if type(bcTfIncidence.index[0]) is str:
            bcTfIncidence = bcTfIncidence.loc[index.astype(str),:]
        elif type(bcTfIncidence.index[0]) is unicode:
            bcTfIncidence = bcTfIncidence.loc[index.astype(unicode),:]
        else:
            bcTfIncidence = bcTfIncidence.loc[index,:]

        return bcTfIncidence

    allBcs = mechanisticOutput.keys()
    allTfs = list(set(numpy.hstack([mechanisticOutput[i].keys() for i in mechanisticOutput.keys()])))

    bcTfIncidence = pandas.DataFrame(numpy.zeros((len(allBcs),len(allTfs))))
    bcTfIncidence.index = allBcs
    bcTfIncidence.columns = allTfs

    for bc in mechanisticOutput.keys():
        bcTfs = mechanisticOutput[bc].keys()
        bcTfIncidence.loc[bc,bcTfs] = 1

    index = numpy.sort(numpy.array(bcTfIncidence.index).astype(int))
    if type(bcTfIncidence.index[0]) is str:
        bcTfIncidence = bcTfIncidence.loc[index.astype(str),:]
    elif type(bcTfIncidence.index[0]) is unicode:
        bcTfIncidence = bcTfIncidence.loc[index.astype(unicode),:]
    else:
        bcTfIncidence = bcTfIncidence.loc[index,:]

    return bcTfIncidence


def __read_pkl(input_file):
    with open(input_file, 'rb') as f:
        return pickle.load(f)


def tf_expression(expressionData,motifPath=os.path.join("..","data","all_tfs_to_motifs.pkl")):
    allTfsToMotifs = __read_pkl(motifPath)
    tfs = list(set(allTfsToMotifs.keys())&set(expressionData.index))
    tfExp = expressionData.loc[tfs,:]
    return tfExp


def __filter_mutations(mutationFile, minNumMutations=None):
    mutations = pandas.read_csv(mutationFile, index_col=0, header=0)
    if minNumMutations is None:
        minNumMutations = min(numpy.ceil(mutations.shape[1] * 0.01), 4)
    freqMuts = list(mutations.index[numpy.where(numpy.sum(mutations,axis=1) >= minNumMutations)[0]])
    return mutations.loc[freqMuts,:]


def __get_mutations(mutationString,mutationMatrix):
    return mutationMatrix.columns[numpy.where(mutationMatrix.loc[mutationString,:]>0)[0]]

def __mutation_regulator_stratification(mutationDf,tfDf,threshold=0.05,dictionary_=False):
    incidence = pandas.DataFrame(numpy.zeros((tfDf.shape[0],mutationDf.shape[0])))
    incidence.index = tfDf.index
    incidence.columns = mutationDf.index

    stratification = {}
    tfCols = set(tfDf.columns)
    mutCols = set(mutationDf.columns)
    for mutation in mutationDf.index:
        mut = __get_mutations(mutation,mutationDf)
        wt = list(mutCols-set(mut))
        mut = list(set(mut)&tfCols)
        wt = list(set(wt)&tfCols)
        tmpMut = tfDf.loc[:,mut]
        tmpWt = tfDf.loc[:,wt]
        ttest = stats.ttest_ind(tmpMut,tmpWt,axis=1,equal_var=False)
        significant = numpy.where(ttest[1]<=threshold)[0]
        hits = list(tfDf.index[significant])
        if len(hits) > 0:
            incidence.loc[hits,mutation] = 1
            if dictionary_ is not False:
                stratification[mutation] = {}
                for i in range(len(hits)):
                    stratification[mutation][hits[i]] = [ttest[0][significant[i]],ttest[1][significant[i]]]

    if dictionary_ is not False:
        return incidence, stratification
    return incidence


def generate_inputs(expressionData,
                    mechanisticOutput,
                    coexpressionModules,
                    saveFolder,
                    dataFolder,
                    mutationFile="filteredMutationsIA12.csv",
                    regulon_dict=None):
    #bcTfIncidence
    bcTfIncidence = _bicluster_tf_incidence(mechanisticOutput,regulons=regulon_dict)
    bcTfIncidence.to_csv(os.path.join(saveFolder,"bcTfIncidence.csv"))

    #eigengenes
    eigengenes = subtypes.principal_df(coexpressionModules, expressionData, subkey=None,
                                       regulons=regulon_dict, minNumberGenes=1)
    eigengenes = eigengenes.T
    index = numpy.sort(numpy.array(eigengenes.index).astype(int))
    eigengenes = eigengenes.loc[index.astype(str),:]
    eigengenes.to_csv(os.path.join(saveFolder,"eigengenes.csv"))

    #tfExpression
    tfExp = tf_expression(expressionData,
                            motifPath=os.path.join(dataFolder, "all_tfs_to_motifs.pkl"))
    tfExp.to_csv(os.path.join(saveFolder,"tfExpression.csv"))

    #filteredMutations:
    filteredMutations = __filter_mutations(mutationFile)
    filteredMutations.to_csv(os.path.join(saveFolder,"filteredMutations.csv"))

    #regStratAll
    tfStratMutations = __mutation_regulator_stratification(filteredMutations, tfDf=tfExp,
                                                           threshold=0.01)
    keepers = list(set(numpy.arange(tfStratMutations.shape[1])) -
                   set(numpy.where(numpy.sum(tfStratMutations, axis=0) == 0)[0]))
    tfStratMutations = tfStratMutations.iloc[:,keepers]
    tfStratMutations.to_csv(os.path.join(saveFolder,"regStratAll.csv"))


def process_causal_results(causalPath=os.path.join("..","results","causal"), causalDictionary=False):
    causalFiles = []
    for root, dirs, files in os.walk(causalPath, topdown=True):
       for name in files:
          if name.split(".")[-1] == 'DS_Store':
              continue
          causalFiles.append(os.path.join(root, name))

    if causalDictionary is False:
        causalDictionary = {}
    for csv in causalFiles:
        tmpcsv = pandas.read_csv(csv,index_col=False,header=None)
        for i in range(1,tmpcsv.shape[0]):
            score = float(tmpcsv.iloc[i,-2])
            if score <1:
                break
            bicluster = int(tmpcsv.iloc[i,-3].split(":")[-1].split("_")[-1])
            if bicluster not in causalDictionary.keys():
                causalDictionary[bicluster] = {}
            regulator = tmpcsv.iloc[i,-5].split(":")[-1]
            if regulator not in causalDictionary[bicluster].keys():
                causalDictionary[bicluster][regulator] = []
            mutation = tmpcsv.iloc[i,1].split(":")[-1]
            if mutation not in causalDictionary[bicluster][regulator]:
                causalDictionary[bicluster][regulator].append(mutation)

    return causalDictionary

def mutation_matrix(mutation_files, minNumMutations=None):
    matrices = []
    for f in mutation_files:
        matrix = __filter_mutations(mutationFile=f, minNumMutations=minNumMutations)
        matrices.append(matrix)
    return pandas.concat(matrices,axis=0)


def analyze_causal_results(task):
    start, stop = task[0]
    preProcessedCausalResults,mechanisticOutput,filteredMutations,tfExp,eigengenes = task[1]
    postProcessed = {}
    if mechanisticOutput is not None:
        mechOutKeyType = type(mechanisticOutput.keys()[0])
    allPatients = set(filteredMutations.columns)
    keys = preProcessedCausalResults.keys()[start:stop]
    ct=-1
    for bc in keys:
        ct+=1
        if ct%10 == 0:
            print(ct)
        postProcessed[bc] = {}
        for tf in preProcessedCausalResults[bc].keys():
            for mutation in preProcessedCausalResults[bc][tf]:
                mut = __get_mutations(mutation,filteredMutations)
                wt = list(allPatients - set(mut))
                mutTfs = tfExp.loc[tf,mut][tfExp.loc[tf,mut]>-4.01]
                if len(mutTfs) <=1:
                    mutRegT = 0
                    mutRegP = 1
                elif len(mutTfs) >1:
                    wtTfs = tfExp.loc[tf,wt][tfExp.loc[tf,wt]>-4.01]
                    mutRegT, mutRegP = stats.ttest_ind(list(mutTfs),list(wtTfs),equal_var=False)
                mutBc = eigengenes.loc[bc,mut][eigengenes.loc[bc,mut]>-4.01]
                if len(mutBc) <=1:
                    mutBcT = 0
                    mutBcP = 1
                    mutCorrR = 0
                    mutCorrP = 1
                elif len(mutBc) >1:
                    wtBc = eigengenes.loc[bc,wt][eigengenes.loc[bc,wt]>-4.01]
                    mutBcT, mutBcP = stats.ttest_ind(list(mutBc),list(wtBc),equal_var=False)
                    if len(mutTfs) <=2:
                        mutCorrR = 0
                        mutCorrP = 1
                    elif len(mutTfs) >2:
                        nonzeroPatients = list(set(numpy.array(mut)[tfExp.loc[tf,mut]>-4.01]) &
                                               set(numpy.array(mut)[eigengenes.loc[bc,mut]>-4.01]))
                        mutCorrR, mutCorrP = stats.pearsonr(list(tfExp.loc[tf,nonzeroPatients]),list(eigengenes.loc[bc,nonzeroPatients]))
                signMutTf = 1
                if mutRegT < 0:
                    signMutTf = -1
                elif mutRegT == 0:
                    signMutTf = 0
                signTfBc = 1
                if mutCorrR < 0:
                    signTfBc = -1
                elif mutCorrR == 0:
                    signTfBc = 0
                if mechanisticOutput is not None:
                    if mechOutKeyType is int:
                        phyper = mechanisticOutput[bc][tf][0]
                    elif mechOutKeyType is not int:
                        phyper = mechanisticOutput[str(bc)][tf][0]
                elif mechanisticOutput is None:
                    phyper = 1e-10
                pMutRegBc = 10**-((-numpy.log10(mutRegP) - numpy.log10(mutBcP) -
                                   numpy.log10(mutCorrP) - numpy.log10(phyper))/4.)

                pWeightedTfBc = 10**-((-numpy.log10(mutCorrP) - numpy.log10(phyper))/2.)
                mutFrequency = len(mut) / float(filteredMutations.shape[1])
                postProcessed[bc][tf] = {}
                postProcessed[bc][tf]["regBcWeightedPValue"] = pWeightedTfBc
                postProcessed[bc][tf]["edgeRegBc"] = signTfBc
                postProcessed[bc][tf]["regBcHyperPValue"] = phyper
                if "mutations" not in postProcessed[bc][tf].keys():
                    postProcessed[bc][tf]["mutations"] = {}
                postProcessed[bc][tf]["mutations"][mutation] = {}
                postProcessed[bc][tf]["mutations"][mutation]["mutationFrequency"] = mutFrequency
                postProcessed[bc][tf]["mutations"][mutation]["mutRegBcWeightedPValue"] = pMutRegBc
                postProcessed[bc][tf]["mutations"][mutation]["edgeMutReg"] = signMutTf
                postProcessed[bc][tf]["mutations"][mutation]["mutRegPValue"] = mutRegP
                postProcessed[bc][tf]["mutations"][mutation]["mutBcPValue"] = mutBcP
                postProcessed[bc][tf]["mutations"][mutation]["regBcCorrPValue"] = mutCorrP
                postProcessed[bc][tf]["mutations"][mutation]["regBcCorrR"] = mutCorrR
    return postProcessed



def post_process_causal_results(preProcessedCausalResults, filteredMutations, tfExp,
                                eigengenes, mechanisticOutput=None, numCores=5):
    taskSplit = util.split_for_multiprocessing(preProcessedCausalResults.keys(),numCores)
    taskData = (preProcessedCausalResults, mechanisticOutput, filteredMutations, tfExp, eigengenes)
    tasks = [[taskSplit[i],taskData] for i in range(len(taskSplit))]
    Output = util.multiprocess(analyze_causal_results, tasks)
    return util.condense_output(Output)


def causal_mechanistic_network_dictionary(postProcessedCausalAnalysis,
                                          biclusterRegulatorPvalue=0.05,
                                          regulatorMutationPvalue=0.05,
                                          mutationFrequency=0.025,
                                          requireCausal=False):
    tabulatedResults = []
    ct=-1

    for key in postProcessedCausalAnalysis.keys():
        ct+=1
        if ct%10==0:
            print(ct)
        lines = []
        regs = postProcessedCausalAnalysis[key].keys()
        for reg in regs:
            bcid = key
            regid = reg
            bcRegEdgeType = int(postProcessedCausalAnalysis[key][reg]['edgeRegBc'])
            bcRegEdgePValue = postProcessedCausalAnalysis[key][reg]['regBcWeightedPValue']
            bcTargetEnrichmentPValue = postProcessedCausalAnalysis[key][reg]['regBcHyperPValue']
            if bcRegEdgePValue <= biclusterRegulatorPvalue:
                if len(postProcessedCausalAnalysis[key][reg]['mutations'])>0:
                    for mut in postProcessedCausalAnalysis[key][reg]['mutations'].keys():
                        mutFrequency = postProcessedCausalAnalysis[key][reg]['mutations'][mut]['mutationFrequency']
                        mutRegPValue = postProcessedCausalAnalysis[key][reg]['mutations'][mut]['mutRegPValue']
                        if mutFrequency >= mutationFrequency:
                            if mutRegPValue <= regulatorMutationPvalue:
                                mutid = mut
                                mutRegEdgeType = int(postProcessedCausalAnalysis[key][reg]['mutations'][mut]['edgeMutReg'])
                            elif mutRegPValue > regulatorMutationPvalue:
                                mutid = numpy.nan #"NA"
                                mutRegEdgeType = numpy.nan #"NA"
                                mutRegPValue = numpy.nan #"NA"
                                mutFrequency = numpy.nan #"NA"
                        elif mutFrequency < mutationFrequency:
                            mutid = numpy.nan #"NA"
                            mutRegEdgeType = numpy.nan #"NA"
                            mutRegPValue = numpy.nan #"NA"
                            mutFrequency = numpy.nan #"NA"
                elif len(postProcessedCausalAnalysis[key][reg]['mutations'])==0:
                    mutid = numpy.nan #"NA"
                    mutRegEdgeType = numpy.nan #"NA"
                    mutRegPValue = numpy.nan #"NA"
                    mutFrequency = numpy.nan #"NA"
            elif bcRegEdgePValue > biclusterRegulatorPvalue:
                continue
            line = [bcid,regid,bcRegEdgeType,bcRegEdgePValue,bcTargetEnrichmentPValue,mutid,mutRegEdgeType,mutRegPValue,mutFrequency]
            lines.append(line)
        if len(lines) == 0:
            continue
        stack = numpy.vstack(lines)
        df = pandas.DataFrame(stack)
        df.columns = ["Cluster","Regulator","RegulatorToClusterEdge","RegulatorToClusterPValue","RegulatorBindingSiteEnrichment","Mutation","MutationToRegulatorEdge","MutationToRegulatorPValue","FrequencyOfMutation"]
        tabulatedResults.append(df)

    resultsDf = pandas.concat(tabulatedResults,axis=0)
    resultsDf = resultsDf[resultsDf["RegulatorToClusterEdge"]!='0']
    resultsDf.index = numpy.arange(resultsDf.shape[0])

    if requireCausal is True:
        resultsDf = resultsDf[resultsDf["Mutation"]!="nan"]

    return resultsDf
